#include <algorithm>
#include <msp/core/raii.h>
#include "timer.h"
#include "utils.h"

using namespace std;

namespace Msp {
namespace Time {

Timer::Timer():
	sem(1),
	blocking(false)
{ }

Timer::~Timer()
{
	for(vector<SlotProxy>::iterator i=slots.begin(); i!=slots.end(); ++i)
		delete i->slot;
}

Timer::Slot &Timer::add(const TimeDelta &td)
{
	Slot *s = new Slot(td);
	MutexLock l(mutex);
	slots.push_back(s);
	push_heap(slots.begin(), slots.end());
	if(blocking)
		sem.signal();
	return *s;
}

Timer::Slot &Timer::add(const TimeStamp &ts)
{
	Slot *s = new Slot(ts);
	MutexLock l(mutex);
	slots.push_back(s);
	push_heap(slots.begin(), slots.end());
	if(blocking)
		sem.signal();
	return *s;
}

void Timer::cancel(Slot &slot)
{
	MutexLock l(mutex);
	for(vector<SlotProxy>::iterator i=slots.begin(); i!=slots.end(); ++i)
		if(i->slot==&slot)
		{
			delete i->slot;
			slots.erase(i);
			make_heap(slots.begin(), slots.end());
			return;
		}
}

void Timer::tick(bool block)
{
	Slot *next = 0;
	{
		MutexLock l(mutex);
		while(1)
		{
			TimeStamp stamp;
			TimeStamp t = now();
			if(!slots.empty())
			{
				next = slots.begin()->slot;
				stamp = next->get_timeout();
				if(stamp<=t)
					break;
			}

			if(block)
			{
				SetFlag setf(blocking);
				mutex.unlock();
				if(stamp)
					sem.wait(stamp-t);
				else
					sem.wait();
				mutex.lock();
				// The slots may have changed while waiting so check again
				continue;
			}
			else
				return;
		}

		pop_heap(slots.begin(), slots.end());
		slots.pop_back();
	}

	try
	{
		if(next->signal_timeout.emit() && next->increment())
		{
			MutexLock l(mutex);
			slots.push_back(next);
			push_heap(slots.begin(), slots.end());
		}
		else
			delete next;
	}
	catch(...)
	{
		delete next;
		throw;
	}
}

TimeStamp Timer::get_next_timeout() const
{
	if(slots.empty())
		return TimeStamp();
	return slots.begin()->slot->get_timeout();
}


Timer::Slot::Slot(const TimeDelta &td):
	interval(td),
	timeout(now()+interval)
{ }

Timer::Slot::Slot(const TimeStamp &ts):
	timeout(ts)
{ }

bool Timer::Slot::increment()
{
	if(!interval)
		return false;
	timeout += interval;
	return true;
}


Timer::SlotProxy::SlotProxy(Slot *s):
	slot(s)
{ }

bool Timer::SlotProxy::operator<(const SlotProxy &sp) const
{
	return slot->get_timeout()>sp.slot->get_timeout();
}

} // namespace Time
} // namespace Msp
